/* $Id: self.c,v 1.15 2001/06/10 22:28:57 mikpe Exp $
 * self.c
 *
 * This test program illustrates how a process may use the
 * Linux x86 Performance-Monitoring Counters interface to
 * monitor its own execution.
 *
 * The library uses mmap() to map the kernel's accumulated counter
 * state into the process' address space.
 * When perfctr_read_state() is called, it uses the RDPMC and RDTSC
 * instructions to get the current register values, then adds
 * these to the base values found in the mapped-in kernel state.
 * The resulting counts are then delivered to the application.
 *
 * Copyright (C) 1999-2001  Mikael Pettersson
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include "libperfctr.h"

static struct vperfctr *self;
static unsigned nrctrs;	/* # of PMCs, does not include TSC */
static unsigned evntsel0;

void print_cpu_features(unsigned cpu_features) /* move to lib? */
{
    static const char * const features[] = { "rdpmc", "rdtsc", "pcint" };
    int fi, comma;

    for(comma = 0, fi = 0; fi < sizeof features / sizeof features[0]; ++fi) {
	unsigned fmask = 1 << fi;
	if( cpu_features & fmask ) {
	    if( comma )
		printf(",");
	    printf("%s", features[fi]);
	    comma = 1;
	}
    }
}

void do_init(void)
{
    struct perfctr_info info;

    self = vperfctr_open();
    if( !self ) {
	perror("vperfctr_open");
	exit(1);
    }
    if( vperfctr_info(self, &info) < 0 ) {
	perror("vperfctr_info");
	exit(1);
    }
    nrctrs = perfctr_cpu_nrctrs(&info);
    if( nrctrs )
	evntsel0 = perfctr_evntsel_num_insns(&info);
    printf("\nPerfCtr Info:\n");
    printf("version\t\t\t%s\n", info.version);
    printf("nrcpus\t\t\t%u\n", info.nrcpus);
    printf("cpu_type\t\t%u (%s)\n", info.cpu_type, perfctr_cpu_name(&info));
    printf("cpu_features\t\t%#x (", info.cpu_features);
    print_cpu_features(info.cpu_features);
    printf(")\n");
    printf("cpu_khz\t\t\t%lu\n", info.cpu_khz);
    printf("nrctrs\t\t\t%u\n", nrctrs);
}

void do_read(void)
{
    struct vperfctr_state state;

    /* vperfctr_read_state() copies the entire state; when sampling,
       use vperfctr_read_tsc() and vperfctr_read_pmc() instead */
    if( vperfctr_read_state(self, &state) < 0 ) {
	 perror("vperfctr_read_state");
	 exit(1);
    }
    printf("\nCurrent Sample:\n");
    printf("status\t\t\t%d\n", state.status);
    if( nrctrs )
	printf("control.evntsel[0]\t0x%08X\n", state.cpu_control.user.evntsel[0]);
    printf("tsc\t\t\t0x%016llX\n", state.sum.tsc);
    if( nrctrs )
	printf("pmc[0]\t\t\t0x%016llX\n", state.sum.pmc[0]);
}

void do_enable(void)
{
    struct vperfctr_control control;

    memset(&control, 0, sizeof control);
    control.cpu_control.tsc_on = 1;
    if( nrctrs ) {
	control.cpu_control.nractrs = 1;
	control.cpu_control.evntsel[0] = evntsel0;
	/* XXX: pmc_map[] fixup here */
    }
    if( vperfctr_control(self, &control) < 0 ) {
	perror("vperfctr_control");
	exit(1);
    }
}

unsigned fac(unsigned n)
{
    return (n < 2) ? 1 : n * fac(n-1);
}

void do_fac(unsigned n)
{
    printf("\nfac(%u) == %u\n", n, fac(n));
}

int main(void)
{
    do_init();
    do_read();
    do_enable();
    do_fac(15);
    do_read();
    return 0;
}
