/* $Id: libperfctr.c,v 1.19 2001/07/26 21:20:56 mikpe Exp $
 * Library interface to Linux x86 Performance-Monitoring Counters.
 *
 * Copyright (C) 1999-2001  Mikael Pettersson
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include "libperfctr.h"

#define PAGE_SIZE	4096

/*
 * Operations on the process' own virtual-mode perfctrs.
 */

struct vperfctr {
    int fd;
    volatile const struct vperfctr_state *kstate;
    unsigned char have_rdpmc;
    unsigned char have_rdtsc;
};

struct vperfctr *vperfctr_open(void)
{
    struct vperfctr *perfctr;
    struct perfctr_info info;
    int isnew;

    perfctr = malloc(sizeof(*perfctr));
    if( !perfctr )
	return NULL;
    isnew = 1;
    perfctr->fd = open("/proc/self/perfctr", O_RDONLY|O_CREAT);
    if( perfctr->fd < 0 ) {
	isnew = 0;
	perfctr->fd = open("/proc/self/perfctr", O_RDONLY);
	if( perfctr->fd < 0 )
	    goto out_perfctr;
    }
    if( ioctl(perfctr->fd, PERFCTR_INFO, &info) != 0 )
	goto out_fd;
    perfctr->have_rdpmc = (info.cpu_features & PERFCTR_FEATURE_RDPMC) != 0;
    perfctr->have_rdtsc = (info.cpu_features & PERFCTR_FEATURE_RDTSC) != 0;
    perfctr->kstate = mmap(NULL, PAGE_SIZE, PROT_READ,
			   MAP_SHARED, perfctr->fd, 0);
    if( perfctr->kstate != MAP_FAILED ) {
	if( perfctr->kstate->magic == VPERFCTR_MAGIC )
	    return perfctr;
	fprintf(stderr, __FILE__ ":" __FUNCTION__
		": perfctr version mismatch, kernel %#x library %#x\n",
		perfctr->kstate->magic, VPERFCTR_MAGIC);
	munmap((void*)perfctr->kstate, PAGE_SIZE);
    }
 out_fd:
    if( isnew )
	vperfctr_unlink(perfctr);
    close(perfctr->fd);
 out_perfctr:
    free(perfctr);
    return NULL;
}

int vperfctr_info(const struct vperfctr *vperfctr, struct perfctr_info *info)
{
    return ioctl(vperfctr->fd, PERFCTR_INFO, info);
}

#define rdtscl(low)	\
	__asm__ __volatile__("rdtsc" : "=a"(low) : : "edx")
#define rdpmcl(ctr,low)	\
	__asm__ __volatile__("rdpmc" : "=a"(low) : "c"(ctr) : "edx")

unsigned long long vperfctr_read_tsc(const struct vperfctr *self)
{
    unsigned long long sum;
    unsigned int start, old_start, now;

    if( self->kstate->status <= 0 )
	return self->kstate->sum.tsc;

    start = self->kstate->start.tsc;
    do {
	old_start = start;
	rdtscl(now);
	sum = self->kstate->sum.tsc;
	start = self->kstate->start.tsc;
    } while( old_start != start );
    return sum + (now - start);
}

unsigned long long vperfctr_read_pmc(const struct vperfctr *self, unsigned i)
{
    unsigned long long sum;
    unsigned int start, now;
    unsigned int tsc0, tsc1;

    if( self->kstate->status <= 0 )
	return self->kstate->sum.pmc[i];

    if( !self->have_rdpmc || !self->kstate->cpu_control.user.tsc_on ) {
	ioctl(self->fd, VPERFCTR_SAMPLE, NULL);
	return self->kstate->sum.pmc[i];
    }

    tsc1 = self->kstate->start.tsc;
    do {
	tsc0 = tsc1;
	rdpmcl(self->kstate->cpu_control.user.pmc_map[i], now);
	sum = self->kstate->sum.pmc[i];
	start = self->kstate->start.pmc[i];
	tsc1 = self->kstate->start.tsc;
    } while( tsc0 != tsc1 );
    return sum + (now - start);
}

int vperfctr_read_state(const struct vperfctr *self, struct vperfctr_state *state)
{
    unsigned int prev_tsc, next_tsc;
    struct perfctr_low_ctrs now;
    int i, nrctrs;

    /* Fast path is impossible if the counters are stopped, if the TSC
       isn't being sampled (bad idea), or if the CPU doesn't have RDPMC
       and at least one PMC is enabled. */
    nrctrs = self->kstate->cpu_control.user.nractrs;
    if( self->kstate->status <= 0 ||
	!self->kstate->cpu_control.user.tsc_on ||
	(!self->have_rdpmc && nrctrs) ) {
	ioctl(self->fd, VPERFCTR_SAMPLE, NULL);
	do {
	    /* XXX: this copies more than necessary */
	    *state = *self->kstate;
	} while( state->start.tsc != self->kstate->start.tsc );
	return 0;
    }
    /* initial context-switch timestamp */
    next_tsc = self->kstate->start.tsc;
    do {
	prev_tsc = next_tsc;
	/* read register contents */
	if( self->have_rdtsc )	/* WinChip lossage */
	    rdtscl(now.tsc);
	for(i = nrctrs; --i >= 0;)
	    rdpmcl(self->kstate->cpu_control.user.pmc_map[i], now.pmc[i]);
	/* copy kernel state (XXX: too much!) */
	*state = *self->kstate;
	/* next context-switch timestamp */
	next_tsc = self->kstate->start.tsc;
	/* loop until the copy is consistent (the time-stamps match) */
    } while( prev_tsc != next_tsc );
    /* we have a good copy: update the sums */
    if( !self->have_rdtsc )	/* WinChip lossage */
	now.tsc = 0;
    state->sum.tsc += now.tsc - state->start.tsc;
    for(i = nrctrs; --i >= 0;)
	state->sum.pmc[i] += now.pmc[i] - state->start.pmc[i];
    return 0;
}

int vperfctr_control(const struct vperfctr *perfctr,
		     struct vperfctr_control *control)
{
    return ioctl(perfctr->fd, VPERFCTR_CONTROL, control);
}

int vperfctr_stop(const struct vperfctr *perfctr)
{
    return ioctl(perfctr->fd, VPERFCTR_STOP, NULL);
}

int vperfctr_unlink(const struct vperfctr *perfctr)
{
    return ioctl(perfctr->fd, VPERFCTR_UNLINK, NULL);
}

void vperfctr_close(struct vperfctr *perfctr)
{
    munmap((void*)perfctr->kstate, PAGE_SIZE);
    close(perfctr->fd);
    free(perfctr);
}

/*
 * Operations on global-mode perfctrs.
 */

struct gperfctr {
    int fd;
    struct perfctr_info info;	/* XXX: wasteful? */
};

struct gperfctr *gperfctr_open(void)
{
    struct gperfctr *gperfctr;

    gperfctr = malloc(sizeof(*gperfctr));
    if( !gperfctr )
	return NULL;
    gperfctr->fd = open("/dev/perfctr", O_RDONLY);
    if( gperfctr->fd >= 0 ) {
	if( ioctl(gperfctr->fd, PERFCTR_INFO, &gperfctr->info) == 0 )
	    return gperfctr;
	close(gperfctr->fd);
    }
    free(gperfctr);
    return NULL;
}

void gperfctr_close(struct gperfctr *gperfctr)
{
    close(gperfctr->fd);
    free(gperfctr);
}

int gperfctr_control(const struct gperfctr *gperfctr,
		     struct gperfctr_control *arg)
{
    return ioctl(gperfctr->fd, GPERFCTR_CONTROL, (long)arg);
}

int gperfctr_read(const struct gperfctr *gperfctr, struct gperfctr_state *arg)
{
    return ioctl(gperfctr->fd, GPERFCTR_READ, (long)arg);
}

int gperfctr_stop(const struct gperfctr *gperfctr)
{
    return ioctl(gperfctr->fd, GPERFCTR_STOP, 0);
}

int gperfctr_info(const struct gperfctr *gperfctr, struct perfctr_info *info)
{
    *info = gperfctr->info;
    return 0;
}

/*
 * Miscellaneous operations.
 */

unsigned perfctr_cpu_nrctrs(const struct perfctr_info *info)
{
    switch( info->cpu_type ) {
      case PERFCTR_X86_GENERIC:
	return 0;
      case PERFCTR_X86_VIA_C3:
	return 1;
      case PERFCTR_X86_AMD_K7:
	return 4;
      default:
	return 2;
    }
}

const char *perfctr_cpu_name(const struct perfctr_info *info)
{
    switch( info->cpu_type ) {
      case PERFCTR_X86_GENERIC:
	return "Generic x86 with TSC";
      case PERFCTR_X86_INTEL_P5:
        return "Intel Pentium";
      case PERFCTR_X86_INTEL_P5MMX:
        return "Intel Pentium MMX";
      case PERFCTR_X86_INTEL_P6:
        return "Intel Pentium Pro";
      case PERFCTR_X86_INTEL_PII:
        return "Intel Pentium II";
      case PERFCTR_X86_INTEL_PIII:
        return "Intel Pentium III";
      case PERFCTR_X86_CYRIX_MII:
        return "Cyrix 6x86MX/MII/III";
      case PERFCTR_X86_WINCHIP_C6:
	return "WinChip C6";
      case PERFCTR_X86_WINCHIP_2:
	return "WinChip 2/3";
      case PERFCTR_X86_AMD_K7:
	return "AMD K7";
      case PERFCTR_X86_VIA_C3:
	 return "VIA Cyrix III / C3";
      default:
        return "?";
    }
}

unsigned perfctr_evntsel_num_insns(const struct perfctr_info *info)
{
    /* This is terribly naive. Assumes only one event will be
     * selected, at CPL > 0.
     */
    switch( info->cpu_type ) {
      case PERFCTR_X86_INTEL_P5:
      case PERFCTR_X86_INTEL_P5MMX:
      case PERFCTR_X86_CYRIX_MII:
        /* event 0x16, count at CPL 3 */
        return 0x16 | (2 << 6);
      case PERFCTR_X86_INTEL_P6:
      case PERFCTR_X86_INTEL_PII:
      case PERFCTR_X86_INTEL_PIII:
      case PERFCTR_X86_AMD_K7:
        /* event 0xC0, count at CPL > 0, ENable */
        return 0xC0 | (1 << 16) | (1 << 22);
      case PERFCTR_X86_WINCHIP_C6:
        return 0x02;
      case PERFCTR_X86_WINCHIP_2:
        return 0x16;
      case PERFCTR_X86_VIA_C3:
	return 0xC0;
      default:
	return 0;
    }
}
